@@ -1,5 +1,8 @@
 """
-This module is a Numba implementation of the Fortran program surf96.
+Numba implementation of the Fortran program surf96.
+
+This module is not a one-to-one translation from Fortran to Python.
+The code has been adapted and optimized for Numba.
 
 ..
 
@@ -161,7 +164,7 @@ def var(p, q, ra, rb, wvno, xka, xkb, dpth):
 
 
 @jitted
-def dltar1(wvno, omega, d, a, b, rho):
+def dltar1(wvno, omega, d, a, b, rho, llw):
     """Love wave period equation."""
     beta1 = b[-1]
     rho1 = rho[-1]
@@ -172,7 +175,7 @@ def dltar1(wvno, omega, d, a, b, rho):
     e1 = rho1 * rb
     e2 = 1.0 / (beta1 * beta1)
 
-    for m in range(len(d) - 2, -1, -1):
+    for m in range(len(d) - 2, llw, -1):
         beta1 = b[m]
         rho1 = rho[m]
         xmu = rho1 * beta1 * beta1
@@ -212,7 +215,7 @@ def dltar1(wvno, omega, d, a, b, rho):
 
 
 @jitted
-def dltar4(wvno, omega, d, a, b, rho):
+def dltar4(wvno, omega, d, a, b, rho, llw):
     """Rayleigh wave period equation."""
     e = numpy.zeros(5, dtype=numpy.float64)
     ee = numpy.zeros(5, dtype=numpy.float64)
@@ -241,7 +244,7 @@ def dltar4(wvno, omega, d, a, b, rho):
     e[4] = wvno2 - ra * rb
 
     # Matrix multiplication from bottom layer upward
-    for m in range(len(d) - 2, -1, -1):
+    for m in range(len(d) - 2, llw, -1):
         xka = omega / a[m]
         xkb = omega / b[m]
         t = b[m] / omega
@@ -258,7 +261,6 @@ def dltar4(wvno, omega, d, a, b, rho):
         rho1 = rho[m]
         p = ra * dpth
         q = rb * dpth
-        # beta = b[m]
 
         # Evaluate cosP, cosQ...
         _, _, a0, cpcq, cpy, cpz, cqw, cqx, xy, xz, wy, wz = var(
@@ -277,23 +279,36 @@ def dltar4(wvno, omega, d, a, b, rho):
         for i in range(5):
             e[i] = ee[i]
 
-    return e[0]
+    if llw == 0:
+        xka = omega / a[0]
+        wvnop = wvno + xka
+        wvnom = numpy.abs(wvno - xka)
+        ra = numpy.sqrt(wvnop * wvnom)
+        dpth = d[0]
+        rho1 = rho[0]
+        p = ra * dpth
+        w, cosp, _, _, _, _, _, _, _, _, _, _ = var(
+            p, q, ra, 1.0e-5, wvno, xka, xkb, dpth
+        )
+        dlt = cosp * e[0] - rho1 * w * e[1]
+    else:
+        dlt = e[0]
+
+    return dlt
 
 
 @jitted
-def fast_delta(wvno, omega, d, alpha, beta, rho):
+def fast_delta(wvno, omega, d, alpha, beta, rho, llw):
     """
     Fast delta matrix.
 
     After Buchen and Ben-Hador (1996).
 
     """
-    # Convert to meters
-    k = wvno / 1.0e3
-    alpha = alpha * 1.0e3
-    beta = beta * 1.0e3
-    rho = rho * 1.0e3
-    d = d * 1.0e3
+    # Handle water layer
+    if llw == 0:
+        beta = beta * 1.0
+        beta[0] = 1.0e-8
 
     # Initialize arrays
     nl = len(alpha)
@@ -304,10 +319,10 @@ def fast_delta(wvno, omega, d, alpha, beta, rho):
 
     r = numpy.zeros(nl, dtype=numpy.complex_)
     s = numpy.zeros(nl, dtype=numpy.complex_)
-    Ca = numpy.ones(nl, dtype=numpy.complex_)
-    Cb = numpy.ones(nl, dtype=numpy.complex_)
-    Sa = numpy.zeros(nl, dtype=numpy.complex_)
-    Sb = numpy.zeros(nl, dtype=numpy.complex_)
+    Ca = numpy.ones(nl - 1, dtype=numpy.complex_)
+    Cb = numpy.ones(nl - 1, dtype=numpy.complex_)
+    Sa = numpy.zeros(nl - 1, dtype=numpy.complex_)
+    Sb = numpy.zeros(nl - 1, dtype=numpy.complex_)
 
     eps = numpy.zeros(nl - 1, dtype=numpy.complex_)
     eta = numpy.zeros(nl - 1, dtype=numpy.complex_)
@@ -315,19 +330,30 @@ def fast_delta(wvno, omega, d, alpha, beta, rho):
     ap = numpy.zeros(nl - 1, dtype=numpy.complex_)
     b = numpy.zeros(nl - 1, dtype=numpy.complex_)
     bp = numpy.zeros(nl - 1, dtype=numpy.complex_)
+    scale = numpy.zeros(nl - 1, dtype=numpy.int32)
 
     X = numpy.zeros(5, dtype=numpy.complex_)
 
     # Phase velocity
-    c = omega / k
+    c = omega / wvno
     c2 = c * c
 
-    # Other variables
+    # Layer eigenfunctions and other variables
     for i in range(nl):
         mu[i] = rho[i] * beta[i] ** 2
         gam[i] = beta[i] ** 2 / c2
         t[i] = 2.0 - c2 / beta[i] ** 2
 
+        if c < alpha[i]:
+            r[i] = numpy.sqrt(1.0 - c2 / alpha[i] ** 2)
+        elif c > alpha[i]:
+            r[i] = numpy.sqrt(c2 / alpha[i] ** 2 - 1.0) * 1j
+
+        if c < beta[i]:
+            s[i] = numpy.sqrt(1.0 - c2 / beta[i] ** 2)
+        elif c > beta[i]:
+            s[i] = numpy.sqrt(c2 / beta[i] ** 2 - 1.0) * 1j
+
     for i in range(nl - 1):
         eps[i] = rho[i + 1] / rho[i]
         eta[i] = 2.0 * (gam[i] - eps[i] * gam[i + 1])
@@ -336,78 +362,103 @@ def fast_delta(wvno, omega, d, alpha, beta, rho):
         b[i] = 1.0 - eta[i]
         bp[i] = b[i] - 1.0
 
-    # Layer eigenfunctions
-    for i in range(nl):
         if c < alpha[i]:
-            r[i] = numpy.sqrt(1.0 - c2 / alpha[i] ** 2)
-            Ca[i] = numpy.cosh(k * r[i] * d[i])
-            Sa[i] = numpy.sinh(k * r[i] * d[i])
+            Ca[i] = numpy.cosh(wvno * r[i] * d[i])
+            Sa[i] = numpy.sinh(wvno * r[i] * d[i])
         elif c > alpha[i]:
-            r[i] = numpy.sqrt(c2 / alpha[i] ** 2 - 1.0) * 1j
-            Ca[i] = numpy.cos(k * r[i].imag * d[i])
-            Sa[i] = numpy.sin(k * r[i].imag * d[i]) * 1j
+            Ca[i] = numpy.cos(wvno * r[i].imag * d[i])
+            Sa[i] = numpy.sin(wvno * r[i].imag * d[i]) * 1j
 
         if c < beta[i]:
-            s[i] = numpy.sqrt(1.0 - c2 / beta[i] ** 2)
-            Cb[i] = numpy.cosh(k * s[i] * d[i])
-            Sb[i] = numpy.sinh(k * s[i] * d[i])
+            Cb[i] = numpy.cosh(wvno * s[i] * d[i])
+            Sb[i] = numpy.sinh(wvno * s[i] * d[i])
+
+            # Handle hyperbolic overflow
+            if wvno * s[i].real * d[i] > 80.0:
+                scale[i] = 1
+                Cb[i] /= Ca[i]
+                Sb[i] /= Ca[i]
+
         elif c > beta[i]:
-            s[i] = numpy.sqrt(c2 / beta[i] ** 2 - 1.0) * 1j
-            Cb[i] = numpy.cos(k * s[i].imag * d[i])
-            Sb[i] = numpy.sin(k * s[i].imag * d[i]) * 1j
+            Cb[i] = numpy.cos(wvno * s[i].imag * d[i])
+            Sb[i] = numpy.sin(wvno * s[i].imag * d[i]) * 1j
+
+    # Rayleigh-wave fast delta matrix
+    if llw == 0:
+        Cb[0] = 1.0
+        Sb[0] = 0.0
 
-    # Rayleigh-wave fast Delta matrix
     X[0] = 2.0 * t[0]
     X[1] = -t[0] * t[0]
     X[4] = -4.0
     X *= mu[0] * mu[0]
+    normc(X)
 
     for i in range(nl - 1):
         p1 = Cb[i] * X[1] + s[i] * Sb[i] * X[2]
         p2 = Cb[i] * X[3] + s[i] * Sb[i] * X[4]
+        p3 = Cb[i] * X[2]
+        p4 = Cb[i] * X[4]
         if c == beta[i]:
-            p3 = k * d[i] * X[1] + Cb[i] * X[2]
-            p4 = k * d[i] * X[3] + Cb[i] * X[4]
+            p3 += wvno * d[i] * X[1]
+            p4 += wvno * d[i] * X[3]
         else:
-            p3 = Sb[i] * X[1] / s[i] + Cb[i] * X[2]
-            p4 = Sb[i] * X[3] / s[i] + Cb[i] * X[4]
+            p3 += Sb[i] * X[1] / s[i]
+            p4 += Sb[i] * X[3] / s[i]
 
-        q1 = Ca[i] * p1 - r[i] * Sa[i] * p2
-        q3 = Ca[i] * p3 - r[i] * Sa[i] * p4
-        if c == alpha[i]:
-            q2 = -k * d[i] * p3 + Ca[i] * p4
-            q4 = -k * d[i] * p1 + Ca[i] * p2
+        if scale[i] == 1:
+            q1 = p1 - r[i] * p2
+            q3 = p3 - r[i] * p4
+            q2 = p4
+            q4 = p2
+            if c != alpha[i]:
+                q2 -= p3 / r[i]
+                q4 -= p1 / r[i]
         else:
-            q2 = -Sa[i] * p3 / r[i] + Ca[i] * p4
-            q4 = -Sa[i] * p1 / r[i] + Ca[i] * p2
+            q1 = Ca[i] * p1 - r[i] * Sa[i] * p2
+            q3 = Ca[i] * p3 - r[i] * Sa[i] * p4
+            q2 = Ca[i] * p4
+            q4 = Ca[i] * p2
+            if c == alpha[i]:
+                q2 -= wvno * d[i] * p3
+                q4 -= wvno * d[i] * p1
+            else:
+                q2 -= Sa[i] * p3 / r[i]
+                q4 -= Sa[i] * p1 / r[i]
 
-        y1 = ap[i] * X[0] + a[i] * q1
-        y2 = a[i] * X[0] + ap[i] * q2
-        z1 = b[i] * X[0] + bp[i] * q1
-        z2 = bp[i] * X[0] + b[i] * q2
+        y1 = a[i] * q1
+        y2 = ap[i] * q2
+        z1 = bp[i] * q1
+        z2 = b[i] * q2
+        if scale[i] == 0:
+            y1 += ap[i] * X[0]
+            y2 += a[i] * X[0]
+            z1 += b[i] * X[0]
+            z2 += bp[i] * X[0]
 
         X[0] = bp[i] * y1 + b[i] * y2
         X[1] = a[i] * y1 + ap[i] * y2
         X[2] = eps[i] * q3
         X[3] = eps[i] * q4
         X[4] = bp[i] * z1 + b[i] * z2
+        normc(X)
 
     return numpy.real(X[1] + s[-1] * X[3] - r[-1] * (X[3] + s[-1] * X[4]))
 
 
 @jitted
-def dltar(wvno, omega, d, a, b, rho, ifunc):
+def dltar(wvno, omega, d, a, b, rho, ifunc, llw):
     """Select Rayleigh or Love wave period equation."""
     if ifunc == 1:
-        return dltar1(wvno, omega, d, a, b, rho)
+        return dltar1(wvno, omega, d, a, b, rho, llw)
     elif ifunc == 2:
-        return dltar4(wvno, omega, d, a, b, rho)
+        return dltar4(wvno, omega, d, a, b, rho, llw)
     else:
-        return fast_delta(wvno, omega, d, a, b, rho)
+        return fast_delta(wvno, omega, d, a, b, rho, llw)
 
 
 @jitted
-def nevill(t, c1, c2, del1, del2, d, a, b, rho, ifunc):
+def nevill(t, c1, c2, del1, del2, d, a, b, rho, ifunc, llw):
     """Hybrid method for refining root once it has been bracketted."""
     x = numpy.zeros(20, dtype=numpy.float64)
     y = numpy.zeros(20, dtype=numpy.float64)
@@ -415,7 +466,7 @@ def nevill(t, c1, c2, del1, del2, d, a, b, rho, ifunc):
     # Initial guess
     omega = twopi / t
     c3 = 0.5 * (c1 + c2)
-    del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc)
+    del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc, llw)
     nev = 1
     nctrl = 1
 
@@ -429,7 +480,7 @@ def nevill(t, c1, c2, del1, del2, d, a, b, rho, ifunc):
         if c3 < min(c1, c2) or c3 > max(c1, c2):
             nev = 0
             c3 = 0.5 * (c1 + c2)
-            del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc)
+            del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc, llw)
 
         s13 = del1 - del3
         s32 = del3 - del2
@@ -459,7 +510,7 @@ def nevill(t, c1, c2, del1, del2, d, a, b, rho, ifunc):
         s2 = 0.01 * ss2
         if s1 > ss2 or s2 > ss1 or nev == 0:
             c3 = 0.5 * (c1 + c2)
-            del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc)
+            del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc, llw)
             nev = 1
             m = 1
         else:
@@ -486,13 +537,13 @@ def nevill(t, c1, c2, del1, del2, d, a, b, rho, ifunc):
 
             if flag:
                 c3 = x[0]
-                del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc)
+                del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc, llw)
                 nev = 2
                 m += 1
                 m = min(m, 10)
             else:
                 c3 = 0.5 * (c1 + c2)
-                del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc)
+                del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc, llw)
                 nev = 1
                 m = 1
 
@@ -500,11 +551,11 @@ def nevill(t, c1, c2, del1, del2, d, a, b, rho, ifunc):
 
 
 @jitted
-def getsol(t1, c1, clow, dc, cm, betmx, ifirst, del1st, d, a, b, rho, ifunc):
+def getsol(t1, c1, clow, dc, cm, betmx, ifirst, del1st, d, a, b, rho, ifunc, llw):
     """Bracket dispersion curve and then refine it."""
     # Bracket solution
     omega = twopi / t1
-    del1 = dltar(omega / c1, omega, d, a, b, rho, ifunc)
+    del1 = dltar(omega / c1, omega, d, a, b, rho, ifunc, llw)
     del1st = del1 if ifirst else del1st
     idir = -1.0 if not ifirst and numpy.sign(del1st) * numpy.sign(del1) < 0.0 else 1.0
 
@@ -517,10 +568,10 @@ def getsol(t1, c1, clow, dc, cm, betmx, ifirst, del1st, d, a, b, rho, ifunc):
             c1 = clow
         else:
             omega = twopi / t1
-            del2 = dltar(omega / c2, omega, d, a, b, rho, ifunc)
+            del2 = dltar(omega / c2, omega, d, a, b, rho, ifunc, llw)
 
             if numpy.sign(del1) != numpy.sign(del2):
-                c1 = nevill(t1, c1, c2, del1, del2, d, a, b, rho, ifunc)
+                c1 = nevill(t1, c1, c2, del1, del2, d, a, b, rho, ifunc, llw)
                 iret = c1 > betmx
                 break
 
@@ -564,6 +615,9 @@ def surf96(t, d, a, b, rho, mode, ifunc, dc):
     c = numpy.zeros(kmax, dtype=numpy.float64)
     cg = numpy.zeros(kmax, dtype=numpy.float64)
 
+    # Check for water layer
+    llw = 0 if b[0] <= 0.0 else -1
+
     # Find the extremal velocities to assist in starting search
     betmx = -1.0e20
     betmn = 1.0e20
@@ -587,7 +641,7 @@ def surf96(t, d, a, b, rho, mode, ifunc, dc):
 
     one = 1.0e-2
     onea = 1.5
-    for iq in range(mode):
+    for iq in range(mode + 1):
         ibeg = 0
         iend = kmax
 
@@ -610,14 +664,14 @@ def surf96(t, d, a, b, rho, mode, ifunc, dc):
 
             # Bracket root and refine it
             c1, del1st, iret = getsol(
-                t[k], c1, clow, dc, cm, betmx, ifirst, del1st, d, a, b, rho, ifunc
+                t[k], c1, clow, dc, cm, betmx, ifirst, del1st, d, a, b, rho, ifunc, llw,
             )
 
             if iret and iq > 0:
                 for i in range(k, kmax):
                     cg[i] = 0.0
 
-                if iq == mode - 1:
+                if iq == mode:
                     return cg
                 else:
                     c1 = 0.0
